/* *****************************************************************
 * High Speed Serial Driver for the Dynamic Engineering PMC BiSerial
 * This driver only ddoes transmit... ie implement write  system call
 * This driver is Copyright David Armstrong 2000, 2001
 * This source code is released under GPL see the file gpl.txt for 
 * details
 *******************************************************************
 */ 

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>


#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif


#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>     /* everything... */
#include <linux/proc_fs.h>
#include <linux/errno.h>  /* error codes */
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/tqueue.h>
#include <linux/poll.h>   /* poll table support */
#include <linux/ioctl.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/segment.h>

#define MAX_CARDS 2


#include "hss.h"
#include "hss_it.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#define V2_4 1
#endif

static int order = BUFFER_SIZE_ORDER;
MODULE_PARM(order,"i");
MODULE_PARM_DESC(order,"The amount of kernel spaced buffer to allocate 2^order bytes");

struct Hss
{
#ifndef V2_4
        struct wait_queue *Tx_queue;
#else /* V2_4 */
        wait_queue_head_t Tx_queue;
#endif /* V2_4 */
        unsigned long pci_addr;   /* address and length of the mem spcae used by the hss */
        unsigned long pci_len;
        struct PMC *pPMC;
        Native * Buffer;
        Native * WriteData;      /* pointer to the buffer location to write to */
        Native * ReadData;       /* pointer to the buffer location to read from */
        int Busy;
        int Error;
        int Started;
        int Threshold;           /* The almost empty setpoint in shorts */
        int FifoSize;
        int MaxWrite;
        int Count1;
        int Count2;
        unsigned int irq ;
        int min;
        int max;
        int NuInterrupts;
        long SumTime;
        long SumTimeSqr;
        struct timeval last_time;  /* for debug output */
        struct timeval last_write; /* for debug output */
};



static struct Hss Hss_tbl[MAX_CARDS];
static int hss_major = 0;     /* Driver major number   */
static int nu_cards = 0;      /* number of cards found */

inline Native HssStat0(struct Hss * hs)
{
    return hs->pPMC->Stat0;
}

void   HssEnableIrq(struct Hss *hs)
{
    Native status;

    printk("W Enabling Interrupts\n");
    hs->pPMC->Intc = 1;  /* Write anything to clear the interrupts */
    /* set the master interupt enable */
    status = (hs->pPMC->Base)| INT_EN_MASTER;
    hs->pPMC->Base = status;
    /* Enable the almost empty interrupt and start the TX */
    status = hs->pPMC->Tx;
    status = status | (CLOCK_EN| INT_EN_TX | INT_EN_PAE | STRT_TX);
    hs->pPMC->Tx = status;
}

void HssReset(struct Hss *hs)
{
    // put the fifo reset code around the rest so the pulse is long enough
    hs->pPMC->Base =  FIFO_EN | CLOCK_PCI;   /* clr FIFO hi, test mode */
    udelay(1);
    // Return Reset high
    hs->pPMC->Base =  FIFO_EN | CLOCK_PCI;   /* clr FIFO hi, test mode */
    udelay(1);
    hs->pPMC->Base =  CLOCK_PCI;             /* clr FIFO lo, test mode  */
    hs->pPMC->Intc = 1;  /* Write anything to clear the interrupts */
    hs->pPMC->Base = 0;
    hs->pPMC->Tx = 0;
    hs->pPMC->Rx = 0;
}


void HssDisableIrq(struct Hss * hs)
{
   Native status;

   printk("D Disabling Interrupts\n"); 
   hs->pPMC->Intc = 1;  /* Write anything to clear the interrupts */
   status = hs->pPMC->Tx;
   hs->pPMC->Tx = status & ~( INT_EN_PAE | STRT_TX | CLOCK_EN);
   /* reset the master interupt enable on the PMC  */
   status = hs->pPMC->Base;
//   hs->pPMC->Base = status & ~ (INT_SET | INT_EN_MASTER);
   hs->pPMC->Base = 0;
}

int HssRelease (struct inode *inode, struct file *filp)
{
    unsigned int minor = MINOR(inode->i_rdev);
    struct timeval tv;
    struct Hss *hs = &Hss_tbl[minor];
    if (minor >= nu_cards)
      return -ENXIO;
    if (hs->Busy !=1)
    {
        printk("Device Hss%d was not open\n",minor);
        return -ENXIO;
    }
    HssDisableIrq(hs);
    hs->Busy = 0;
    hs->Error = 0;
    hs->Started = 0;
    hs->WriteData = hs->Buffer;
    hs->ReadData = hs->Buffer;
    HssReset(hs);
    do_gettimeofday( &tv );
    printk("Releasing device at %08lu.%06lu\n",tv.tv_sec,tv.tv_usec );
    MOD_DEC_USE_COUNT;
    return 0;
}

int HssOpen (struct inode *inode, struct file *filp)
{
    unsigned int minor = MINOR(inode->i_rdev);
    struct Hss *hs = &Hss_tbl[minor];
    printk("O Opening device Hss%d\n",minor);
    if (minor >= MAX_CARDS)
    {
        printk("O Invalid minor number %d\n",minor);
        return -ENXIO;
    }
    if (hs->Busy ==1)
    {
        printk("O Device Hss%d is busy\n",minor);
        return -ENXIO;
    }
    hs->Busy = 1;
    hs->Error = 0;
    hs->Started = 0;
    hs->WriteData = hs->Buffer;
    hs->ReadData = hs->Buffer;
    hs->min = 1000000;
    hs->max = 0;
    hs->NuInterrupts = 0;
    hs->SumTime = 0;
    hs->SumTimeSqr = 0;
    HssReset(hs);
    MOD_INC_USE_COUNT;
    return 0;
}


unsigned int HssPoll (struct file *filp, poll_table *table)
{
    return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;
}


ssize_t HssWrite(struct file * file, const char *buf, size_t count, loff_t * ppos)
{
    unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
    int i;
    Native status;
    int Count;
    int SpaceAvailable;
    Native * Temp;
    struct timeval tv;
    const int MaxFifosize = 32768;
    struct Hss *hs = &Hss_tbl[minor];
    /* make sure this is being called with a long (0x3) array */
    if ((count & (sizeof(Native)-1 ))!= 0)
    {
        printk("W Input must be a multiple of the Native word size in bytes: %d\n",count);
        return -EIO;
    }
    Count = count /sizeof(Native);

    if (hs->Error == 1)
    {
        printk("W Error detected refusing write!\n");
        return -EIO;
    }
    /* If not started then just put data into the buffer */
    if (hs->Started == 0)
    {
        /* since we are not started the read pointer == buffer pointer */
        SpaceAvailable = BUFFER_SIZE - (hs->WriteData - hs->Buffer);
        if (Count > SpaceAvailable)
        {
            printk("W Write of %d too large for space available %d\n",Count,SpaceAvailable);
            return -ENOSPC;
        }
        Temp = (Native *) buf;
        /*^ put data into buffer */
        for (i = 0; i < Count; i++)
        {
            *(hs->WriteData++) = *(Temp++);
        }
        /*^ if the buffer is over 3/4 full then start */
        if ((hs->WriteData - hs->Buffer)> BUFFER_SIZE/2 ) /* buffer is over 3/4 full */
        {
            for (i=0; i <MaxFifosize+10; i++)
            {
                hs->pPMC->FifoTx = *(hs->ReadData++);
                if (((status=hs->pPMC->Stat0)& FTX_FFN)==0) break;
            }      
            hs->FifoSize = i;
            hs->MaxWrite =  i - hs->Threshold;
            printk("Fifosize of %d detected\n",i);  
            
            hs->Count1 = hs->WriteData - hs->Buffer - i;
            do_gettimeofday( &tv );
            printk("W Starting at %08lu.%06lu\n",tv.tv_sec,tv.tv_usec );
            if (hs->Count1 > hs->MaxWrite) hs->Count1 = hs->MaxWrite ;
            hs->Count2 = 0;
            hs->Started = 1;
            HssEnableIrq(hs);
            hs->last_time.tv_sec = tv.tv_sec;
            hs->last_time.tv_usec = tv.tv_usec;
            hs->last_write.tv_sec = tv.tv_sec;
            hs->last_write.tv_usec = tv.tv_usec;
            
        }
        return count;
    }
    else
    {
       SpaceAvailable = hs->ReadData - hs->WriteData ;
       if (SpaceAvailable <0 ) SpaceAvailable += BUFFER_SIZE;
       /* if not ready to accept data then sleep */
       while ((SpaceAvailable < Count)&&(hs->Error == 0))
       {
           interruptible_sleep_on(&hs->Tx_queue);
           if (signal_pending(current))
           {
              return -1;
           }    
           SpaceAvailable = hs->ReadData - hs->WriteData ;
           if (SpaceAvailable < 0 ) SpaceAvailable += BUFFER_SIZE;
       }
       if (hs->Error == 1)
       {
           printk("W Underrun error detected \n");
           return -EIO;
       }
       do_gettimeofday( &tv );
       hs->last_write.tv_sec = tv.tv_sec;
       hs->last_write.tv_usec = tv.tv_usec;
       Temp = (Native *)buf;
       for (i = 0; i <Count ; i++)
       {
            *(hs->WriteData++) = *(Temp++);
            if (hs->WriteData == (hs->Buffer + BUFFER_SIZE)) hs->WriteData = hs->Buffer;
       }
       return count;

    }
}

int HssIoctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    unsigned int minor = MINOR(inode->i_rdev);
    int retval,offset;
    Native status;
    struct Hss *hs = &Hss_tbl[minor];

    if (minor >= nu_cards)
      return -ENXIO;
    switch (cmd)
    {

        case BiTx_READ:
                status = hs->pPMC->Tx;
                if ( copy_to_user((int *) arg, &status,sizeof(long)))
                  return -EFAULT;
                break;
        case BiTx_WRITE:
                if ( copy_from_user(&status ,(int *) arg,sizeof(long)))
                  return -EFAULT;
                hs->pPMC->Tx = status;
                break;
        case BiRx_READ:
                status = hs->pPMC->Rx;
                if ( copy_to_user((int *) arg, &status,sizeof(long)))
                  return -EFAULT;
                break;
        case BiRx_WRITE:
                if ( copy_from_user(&status ,(int *) arg,sizeof(long)))
                  return -EFAULT;
                hs->pPMC->Rx = status;
                break;
        case BiSTAT0_READ:
                status = hs->pPMC->Stat0;
                if ( copy_to_user((int *) arg, &status,sizeof(long)))
                  return -EFAULT;
                break;
        case BiBASE_READ:
                status = hs->pPMC->Base;
                if ( copy_to_user((int *) arg, &status,sizeof(long)))
                  return -EFAULT;
                break;
        case BiBASE_WRITE:
                if ( copy_from_user(&status ,(int *) arg,sizeof(long)))
                  return -EFAULT;
                hs->pPMC->Base = status;
                break;
        case BiDIR_TERM_WRITE:
                if ( copy_from_user(&status ,(int *) arg,sizeof(long)))
                  return -EFAULT;
                hs->pPMC->DirTerm = status;
                break;

        case BiRESET:
                HssReset(hs);
                break;
        case BiTX_FIFO_SET_AE:
                hs->pPMC->Base =   FRX_LD | FTX_LD| FIFO_EN | CLOCK_PCI;
                udelay(1);
                hs->pPMC->Base =  FRX_LD | FIFO_EN | CLOCK_PCI;   /* clr FIFO hi, LD lo for TX, test mode */
                udelay(1);
                hs->pPMC->Base =  FRX_LD | CLOCK_PCI;             /* clr FIFO lo, LD lo for TX, test mode  */
                udelay(1);
                hs->pPMC->Base =  FRX_LD | FIFO_EN | CLOCK_PCI;   /* clr FIFO hi, LD lo for TX, test mode */

                if ( copy_from_user(&offset ,(int *) arg,sizeof(Native)))
                  return -EFAULT;
                hs->Threshold = offset; /* Since this is the number of bytes that the FIFO PAF irq is set to */
                /* and since there are 4 fifo this is the number of longs */
                /* the FIFO is actually made up of 4 8 bit wide FIFO's */
                /* each FIFO has its owm programmable flags            */
                /* only one output will be driven into the FPGA so.... */
                /* make the all the same */
                hs->pPMC->FifoTx = ((offset&0xff)<<24)+((offset&0xff)<<16)+((offset&0xff)<<8)+(offset&0xff) ;
                hs->pPMC->FifoTx = ((offset &0xff00)<<16)+((offset &0xff00)<<8)+((offset &0xff00))+((offset &0xff00)>>8);
                /* clr FIFO hi, LD lo for TX, test mode */
                hs->pPMC->Base =   FRX_LD | FTX_LD| FIFO_EN | CLOCK_PCI;
                udelay(1);
                break;
        case BiRX_FIFO_SET_AF:
                udelay(1);
                hs->pPMC->Base =   FRX_LD | FTX_LD| FIFO_EN | CLOCK_PCI;
                udelay(1);
                hs->pPMC->Base =  FTX_LD | FIFO_EN |CLOCK_PCI;    /* clr FIFO hi, LD lo for TX, test mode */
                udelay(1);
                hs->pPMC->Base =  FTX_LD | CLOCK_PCI;             /* clr FIFO lo, LD lo for TX, test mode  */
                udelay(1);
                hs->pPMC->Base =  FTX_LD | FIFO_EN |CLOCK_PCI;    /* clr FIFO hi, LD lo for TX, test mode */
                udelay(1);
                if ( copy_from_user(&offset ,(int *) arg,sizeof(Native)))
                  return -EFAULT;

                /* two writes to move past the almost empty flags */
                hs->pPMC->FifoRx = status;
                hs->pPMC->FifoRx = status;

                /* this one is the low byte .. */
                /* the FIFO is actually made up of 4 8 bit wide FIFO's */
                /* each FIFO has its owm programmable flags            */
                /* only one output will be driven into the FPGA so.... */
                /* make the all the same */
                hs->pPMC->FifoRx = ((offset&0xff)<<24)+((offset&0xff)<<16)+((offset&0xff)<<8)+(offset&0xff) ;
                hs->pPMC->FifoRx = ((offset &0xff00)<<16)+((offset &0xff00)<<8)+((offset &0xff00))+((offset &0xff00)>>8);
                /* clr FIFO hi, LD lo for TX, test mode */
                hs->pPMC->Base =   FRX_LD | FTX_LD| FIFO_EN | CLOCK_PCI;
                udelay(1);
                break;
        case BiTX_FIFO_WRITE:
                if ( copy_from_user(&status ,(long *) arg,sizeof(Native)))
                  return -EFAULT;
                hs->pPMC->FifoTx = status;
                break;
        case BiTX_FIFO_READ:
                status =  hs->pPMC->FifoTx ;
                if ( copy_to_user((int *) arg, &status,sizeof(Native)))
                  return -EFAULT;
                break;
        case BiRX_FIFO_WRITE:
                if ( copy_from_user(&status ,(long *) arg,sizeof(Native)))
                  return -EFAULT;
                hs->pPMC->FifoRx = status;
                break;
        case BiRX_FIFO_READ:
                status = hs->pPMC->FifoRx;
                if ( copy_to_user((long *) arg, &status,sizeof(Native)))
                  return -EFAULT;
                break;
        default:
                printk("IO CTL default handler invoked\n");
                printk("  IO CTL cmd 0x%x  arg 0x%lx\n",cmd,arg);
                retval = -EINVAL;
    }
    return 0;
}


struct file_operations hss_fops = {
    write: HssWrite,
    poll: HssPoll,
    ioctl: HssIoctl,
    open: HssOpen,
    release: HssRelease,
};

void HssInterrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    Native int_status;
    int i;
    int WordsAvailable;  /* number of words available to write to fifo*/
    long DTime;  
    long Second;
    long USecond;
    struct Hss *hs;
    struct timeval tv,tvr;
    hs = (struct Hss *)dev_id;

    /* get interrupt status */
    int_status = hs->pPMC->Stat0; /* for speed go direct to hardware */
    /* Normally on a share interrupt driver we would test to see if this interrupt is really for us */
    /* if not we would return, however in interest of speed (for this driver) we will test each one */
    /* if none were true we will fall through */
    if ((int_status &  TX_INT )!=0)
    {
        do_gettimeofday( &tvr );
        /*^ Adjust WordsAvailable  for buffer wrap */
        printk("I Last Write at %08lu.%06lu \n",hs->last_write.tv_sec,hs->last_write.tv_usec );
        printk("I Tx FIFOs completely empty  at %08lu.%06lu\n",tvr.tv_sec,tvr.tv_usec );
        /* Indicate that an error has occured */
        hs->Error = 1;
        HssDisableIrq(hs);
        if (waitqueue_active(&hs->Tx_queue))
             wake_up_interruptible(&hs->Tx_queue); /* awake any writing process */
    }
    if  ((int_status & FAE_INT) != 0)
    {
        do_gettimeofday( &tv );
        /*^ if hs-Count2 is not zero we have a two part fifo filling */
        if ((i= hs->Count2) != 0)
        {
            while (i-- >0)
            {
                hs->pPMC->FifoTx = *(hs->ReadData++);    /* put data into FIFO */
            }
            /*^ if pointer is at end of buffer wrap */
            if (hs->ReadData == ( hs->Buffer + BUFFER_SIZE))  /* wrap buffer */
            {
                hs->ReadData = hs->Buffer;
            }
            /*^ else error */
            else
            {
                printk(" Wrong C2 %d C1 %d at %08lu.%06lu\n",hs->Count2,hs->Count1,tv.tv_sec,  tv.tv_usec );
                hs->Error = 1;
                HssDisableIrq(hs);
            }
            hs->Count2 = 0;
        }
        /*^ do the rest (or all) */
        i = hs->Count1;
        /*^ load the fifo */
        while (i-- >0)
        {
            hs->pPMC->FifoTx =*(hs->ReadData++);
        }
        /*^ test to make sure the read pointer is valid */
        if ((hs->ReadData >= ( hs->Buffer + BUFFER_SIZE))|| (hs->ReadData <  hs->Buffer))
        {
            printk("Read Pointer %p left in wrong place...hs count2 was %d Hs count1 %d at %08lu.%06lu\n",hs->ReadData,hs->Count2,hs->Count1,tv.tv_sec,tv.tv_usec );
            hs->Error = 1;
            HssDisableIrq(hs);
        }
        /*^ Fill the FIFO up */
        while ((((int_status = hs->pPMC->Stat0) & FTX_FFN) != 0  )&& (hs->ReadData  != (hs->WriteData -1)))
        {
            hs->pPMC->FifoTx = *(hs->ReadData++);
            if (hs->ReadData == ( hs->Buffer + BUFFER_SIZE)) hs->ReadData = hs->Buffer;
        }

        /*^ Calculate the number of elements available to be put into FIFO */
        WordsAvailable = hs->WriteData - 1 - hs->ReadData ;

        /*^ Adjust WordsAvailable  for buffer wrap */
        if (WordsAvailable < 0 ) WordsAvailable += BUFFER_SIZE;

        /*^ Set hs->Count1 to the max we can write */
        if (WordsAvailable < hs->MaxWrite)
        {
            hs->Count1 = WordsAvailable;
            printk("I Last Write at  %08lu.%06lu WA %d\n",hs->last_write.tv_sec,hs->last_write.tv_usec,WordsAvailable );
            printk("I Running out at %08lu.%06lu\n",tv.tv_sec,tv.tv_usec);
        }
        else
        {
            hs->Count1 = hs->MaxWrite;
        }
        /*^ If the Read Pointer is too near the end then split into two pieces */
        if ( hs->Buffer + BUFFER_SIZE - hs->ReadData   <= hs->Count1)
        {
           hs->Count2 = hs->Buffer + BUFFER_SIZE - hs->ReadData;  
           hs->Count1 = hs->Count1 - hs->Count2;
        }

        /* if start tx bit is reset then we an error */
        if (((hs->pPMC->Tx) & STRT_TX) == 0)
        {
            hs->Error = 1;
            HssDisableIrq(hs);
            printk("I Start TX reset ... %08lu.%06lu \n",  tv.tv_sec,  tv.tv_usec ); 
        }

        if (waitqueue_active(&hs->Tx_queue))
             wake_up_interruptible(&hs->Tx_queue); /* awake any writing process */
        /* calculate stats */
        hs->NuInterrupts += 1;
        /* Calc time */
        Second = tv.tv_sec - hs->last_time.tv_sec ;
        USecond = tv.tv_usec - hs->last_time.tv_usec ;
        DTime = Second * 1000 + (USecond+5000)/1000;
        hs->SumTime = hs->SumTime + DTime;
        hs->SumTimeSqr += (DTime *DTime);
        if (DTime < hs->min) hs->min = DTime;
        if (DTime > hs->max) hs->max = DTime;
        hs->last_time.tv_sec = tv.tv_sec;
        hs->last_time.tv_usec = tv.tv_usec;
    }

    /* check for RX stuff */
    if ((int_status &  FRX_FULL)== 0)
    {
       printk("I RX completely full interrupt received, disabling\n");
       HssDisableIrq(hs);
    }

    if  ((int_status & FAF_INT) != 0)
    {
       printk("I RX almost full interrupt received, disabling\n");
       HssDisableIrq(hs);
    }
    return; /* we have serviced all of the interrupt source IF ANY */
}

static int HssStats( char * page,
    char ** start,
    off_t offset,
    int count,
    int *eof,
    void *data)
{
    char *out = page;
    int len;
    int i;
    if (offset >0)
        return 0;
    out += sprintf(out,"min max #Int SumTime SumTimeSqr\n");
    for (i=0; i< nu_cards;i++)
    {
        struct Hss *hs = &Hss_tbl[i];
        out += sprintf(out, "%d %d %d %ld %ld \n",hs->min,hs->max,hs->NuInterrupts,hs->SumTime,hs->SumTimeSqr);
    }
    len =out -page ;
    *start = page;
    return len;
}



static int HssFifoSize( char * page,
    char ** start,
    off_t offset,
    int count,
    int *eof,
    void *data)
{
    char *out = page;
    int len;
    int i;
    if (offset >0)
        return 0;
    for (i=0; i< nu_cards;i++)
    {
        struct Hss *hs = &Hss_tbl[i];
        out += sprintf(out, "%d \n",hs->FifoSize);
    }
    len =out -page ;
    *start = page;
    return len;
}
    
static int HssThreshold( char * page,
    char ** start,
    off_t offset,
    int count,
    int *eof,
    void *data)
{
    char *out = page;
    int len;
    int i;
    if (offset >0)
        return 0;
    for (i=0; i< nu_cards;i++)
    {
        struct Hss *hs = &Hss_tbl[i];
        out += sprintf(out, "%d \n",hs->Threshold);
    }
    len =out -page ;
    *start = page;
    return len;
}    
    
void proc_HssCreate(void)
{
    struct proc_dir_entry *root, *ent;
    root = create_proc_entry("hss", S_IFDIR, 0);
    if (!root) return;
    ent = create_proc_entry("stats", 0, root);
    if (!ent) return;
    ent->read_proc  = HssStats;
    ent = create_proc_entry("fifosize", 0, root);
    if (!ent) return;
    ent->read_proc  = HssFifoSize;
    ent = create_proc_entry("threshold", 0, root);
    if (!ent) return;
    ent->read_proc  = HssThreshold;
}     

void ReleaseCard(struct Hss * hs)
{
    free_irq(hs->irq, hs);
    HssDisableIrq(hs);
    printk("C RESETing card \n");
    HssReset(hs);
    printk("C Releasing %d pages of memory\n",1<<order);
      if (hs->Buffer) free_pages( (ulong) hs->Buffer,order);
    iounmap((char *)hs->pPMC);
#ifdef V2_4
    printk("C Releaseing PCI space  0x%lx of length %ld \n",hs->pci_addr, hs->pci_len);
    release_mem_region(hs->pci_addr, hs->pci_len);
    
    //iounmap((char *)hs->pPMC);
    
#endif /* V2_4 */
}

void ReleaseCards(int cards)
{
    int i;
    if (cards == 0) return;
    for (i = 0;i < cards ; i++)
    {
       printk("C Releasing card %i \n",i+1);
       ReleaseCard(&Hss_tbl[i]);
    }
    unregister_chrdev(hss_major, "Hss");

}



int init_module(void)
{
  int i;
  struct pci_dev *pcidev = NULL;
  struct Hss *hs;
  int result;
  long temp;
  printk ("I Init PMC BiSerial Driver \n" );
  if (!pci_present())   /* No PCI bus in this machine! */
  {
      printk("I No PCI bus present \n");
      return -ENODEV;
  }
  for(i = 0;i < MAX_CARDS; i++)
  {
    if ( (pcidev = pci_find_device( VendorId, CardId, pcidev)))
    {
        printk("I Found Card %i \n",i+1);
        nu_cards++;
        hs = &Hss_tbl[i];
        hs->irq =  pcidev->irq;
        result = request_irq (  hs->irq ,HssInterrupt, SA_SHIRQ  ,"Hss",hs );
        if (result <0 )
        {
            printk("I Interrupt %d already in use!\n",hs->irq);
            ReleaseCards(nu_cards);
            return result;
        }
        printk("I Interrupt %d assigned\n",hs->irq);
#ifndef V2_4
        hs->pci_addr = pcidev->base_address[0];
#else /* V2_4 */
        //pci_addr = pcidev->base_address[0];
        hs->pci_addr = pci_resource_start(pcidev,0);
        hs->pci_len = pci_resource_len(pcidev,0);        
#endif /* V2_4 */
        // map the memory starting at pci_addr going for length 0x1ffffff to a place the kernel can find it
#ifndef V2_4
        hs->pPMC = (struct PMC *) ioremap(hs->pci_addr, sizeof(struct PMC));
#else /* V2_4 */
        hs->pPMC = (struct PMC *) ioremap(hs->pci_addr, hs->pci_len );

        if (!request_mem_region (hs->pci_addr,hs->pci_len , pcidev->name)) {
          printk ("no mem resource available, aborting\n");
          ReleaseCards(nu_cards);
          return -ENOMEM;
        }
 
#endif /* V2_4 */
        printk("I PCI:found  at 0x%lx mapped to virtual 0x%lx\n",(ulong)hs->pci_addr,(ulong) hs->pPMC);



        /* get some kernel memory for a buffer, the second parameter is the order of the */
        /* memory we want ie order of 3 = 2^3 == 8 pages */
        order = (order - PAGE_SHIFT > 0) ? order - PAGE_SHIFT : 0;
        /* cast it to the type of pointer that we really want   */
        hs->Buffer = (Native *) __get_free_pages(GFP_KERNEL,order); /* never fails */
        if (hs->Buffer == 0)
        {
             printk("I Memory allocation error\n");
             free_irq(hs->irq, hs);
#ifndef V2_4
             iounmap((char *)hs->pPMC);
#else /* V2_4 */
             //iounmap((char *)hs->pPMC);
#endif /* V2_4 */
             ReleaseCards(nu_cards);
             return -ENOMEM;  /*Return error no memory */
        }
        else
        {
            temp = 1<<(order+PAGE_SHIFT);
            printk("I Allocated %d pages ( %ld bytes) of memory at 0x%p \n",1<<order,temp,hs->Buffer);
        }

#ifndef V2_4
        init_waitqueue(&hs->Tx_queue);
#else /* V2_4 */
        init_waitqueue_head(&hs->Tx_queue);
#endif /* V2_4 */
        hs->Error = 0;
        HssDisableIrq(hs);
        printk("I RESETing card %i \n",i+1);
        hs->pPMC->Intc = 1;  /* Write anything to clear the interrupts */
        HssReset(hs);
        // should only register once for the first card
        if (i == 0)
        {
            /* the only way to get in here is on the first card */
            /* so we only have to release its resources */
            result = register_chrdev ( 0 , "Hss" , &hss_fops) ;
            if (result < 0) {
                printk(KERN_INFO "I Hss: can not get major number\n");
                ReleaseCard(&Hss_tbl[0]);
                return result;
            }
            else
            {
                printk("I Hss major number %i\n",result);
            }
            if (hss_major == 0) hss_major = result; /* dynamic */
       }
     }
  }

  if (nu_cards == 0)
  {
       printk("No PCI card present \n");
       return -ENODEV;
  }
  else
  {
      proc_HssCreate();
  }

  return 0;
}
void proc_HssDestroy(void)
{
    remove_proc_entry("stats",0);
    remove_proc_entry("fifosize", 0);
    remove_proc_entry("threshold", 0);
    remove_proc_entry("hss",0);
}        

void cleanup_module(void)
{
    printk("C Releasing %i cards \n",nu_cards);
    ReleaseCards(nu_cards);
    proc_HssDestroy();
}


























